"""Module with a simple buffer implementation using the memory manager"""
from mman import WindowCursor

import sys

__all__ = ["SlidingWindowMapBuffer"]

class SlidingWindowMapBuffer(object):
	"""A buffer like object which allows direct byte-wise object and slicing into 
	memory of a mapped file. The mapping is controlled by the provided cursor.
	
	The buffer is relative, that is if you map an offset, index 0 will map to the 
	first byte at the offset you used during initialization or begin_access
	
	**Note:** Although this type effectively hides the fact that there are mapped windows
	underneath, it can unfortunately not be used in any non-pure python method which
	needs a buffer or string"""
	__slots__ = (
				'_c',			# our cursor
				'_size', 		# our supposed size
				)
	
	
	def __init__(self, cursor = None, offset = 0, size = sys.maxint, flags = 0):
		"""Initalize the instance to operate on the given cursor.
		:param cursor: if not None, the associated cursor to the file you want to access
			If None, you have call begin_access before using the buffer and provide a cursor 
		:param offset: absolute offset in bytes
		:param size: the total size of the mapping. Defaults to the maximum possible size
			From that point on, the __len__ of the buffer will be the given size or the file size.
			If the size is larger than the mappable area, you can only access the actually available
			area, although the length of the buffer is reported to be your given size.
			Hence it is in your own interest to provide a proper size !
		:param flags: Additional flags to be passed to os.open
		:raise ValueError: if the buffer could not achieve a valid state"""
		self._c = cursor
		if cursor and not self.begin_access(cursor, offset, size, flags):
			raise ValueError("Failed to allocate the buffer - probably the given offset is out of bounds")
		# END handle offset

	def __del__(self):
		self.end_access()
		
	def __len__(self):
		return self._size
		
	def __getitem__(self, i):
		c = self._c
		assert c.is_valid()
		if i < 0:
			i = self._size + i
		if not c.includes_ofs(i):
			c.use_region(i, 1)
		# END handle region usage
		return c.buffer()[i-c.ofs_begin()]
	
	def __getslice__(self, i, j):
		c = self._c
		# fast path, slice fully included - safes a concatenate operation and 
		# should be the default
		assert c.is_valid()
		if i < 0:
			i = self._size + i
		if j == sys.maxint:
			j = self._size
		if j < 0:
			j = self._size + j
		if (c.ofs_begin() <= i) and (j < c.ofs_end()):
			b = c.ofs_begin()
			return c.buffer()[i-b:j-b]
		else:
			l = j-i					# total length
			ofs = i
			# Keeping tokens in a list could possible be faster, but the list
			# overhead outweighs the benefits (tested) !
			md = str()
			while l:
				c.use_region(ofs, l)
				assert c.is_valid()
				d = c.buffer()[:l]
				ofs += len(d)
				l -= len(d)
				md += d
			#END while there are bytes to read
			return md
		# END fast or slow path
	#{ Interface
	
	def begin_access(self, cursor = None, offset = 0, size = sys.maxint, flags = 0):
		"""Call this before the first use of this instance. The method was already
		called by the constructor in case sufficient information was provided.
		
		For more information no the parameters, see the __init__ method
		:param path: if cursor is None the existing one will be used. 
		:return: True if the buffer can be used"""
		if cursor:
			self._c = cursor
		#END update our cursor
		
		# reuse existing cursors if possible
		if self._c is not None and self._c.is_associated():
			res = self._c.use_region(offset, size, flags).is_valid()
			if res:
				# if given size is too large or default, we computer a proper size
				# If its smaller, we assume the combination between offset and size
				# as chosen by the user is correct and use it !
				# If not, the user is in trouble.
				if size > self._c.file_size():
					size = self._c.file_size() - offset
				#END handle size
				self._size = size 
			#END set size
			return res
		# END use our cursor
		return False
		
	def end_access(self):
		"""Call this method once you are done using the instance. It is automatically 
		called on destruction, and should be called just in time to allow system
		resources to be freed.
		
		Once you called end_access, you must call begin access before reusing this instance!"""
		self._size = 0
		if self._c is not None:
			self._c.unuse_region()
		#END unuse region
		
	def cursor(self):
		""":return: the currently set cursor which provides access to the data"""
		return self._c
		
	#}END interface





from lib import TestBase, FileCreator

from smmap.mman import *
from smmap.mman import WindowCursor
from smmap.util import align_to_mmap
from smmap.exc import RegionCollectionError

from random import randint
from time import time
import os
import sys
from copy import copy

class TestMMan(TestBase):
	
	def test_cursor(self):
		fc = FileCreator(self.k_window_test_size, "cursor_test")
		
		man = SlidingWindowMapManager()
		ci = WindowCursor(man)	# invalid cursor
		assert not ci.is_valid()
		assert not ci.is_associated()
		assert ci.size() == 0		# this is cached, so we can query it in invalid state
		
		cv = man.make_cursor(fc.path)
		assert not cv.is_valid()	# no region mapped yet
		assert cv.is_associated()# but it know where to map it from
		assert cv.file_size() == fc.size
		assert cv.path() == fc.path
		
		# copy module
		cio = copy(cv)
		assert not cio.is_valid() and cio.is_associated()
		
		# assign method
		assert not ci.is_associated()
		ci.assign(cv)
		assert not ci.is_valid() and ci.is_associated()
		
		# unuse non-existing region is fine
		cv.unuse_region()
		cv.unuse_region()
		
		# destruction is fine (even multiple times)
		cv._destroy()
		WindowCursor(man)._destroy()
		
	def test_memory_manager(self):
		slide_man = SlidingWindowMapManager()
		static_man = StaticWindowMapManager()
		
		for man in (static_man, slide_man):
			assert man.num_file_handles() == 0
			assert man.num_open_files() == 0
			winsize_cmp_val = 0
			if isinstance(man, StaticWindowMapManager):
				winsize_cmp_val = -1
			#END handle window size
			assert man.window_size() > winsize_cmp_val
			assert man.mapped_memory_size() == 0
			assert man.max_mapped_memory_size() > 0
			
			# collection doesn't raise in 'any' mode
			man._collect_lru_region(0)
			# doesn't raise if we are within the limit
			man._collect_lru_region(10)
			
			# doesn't fail if we overallocate 
			assert man._collect_lru_region(sys.maxint) == 0
			
			# use a region, verify most basic functionality
			fc = FileCreator(self.k_window_test_size, "manager_test")
			fd = os.open(fc.path, os.O_RDONLY)
			for item in (fc.path, fd):
				c = man.make_cursor(item)
				assert c.path_or_fd() is item
				assert c.use_region(10, 10).is_valid()
				assert c.ofs_begin() == 10
				assert c.size() == 10
				assert c.buffer()[:] == open(fc.path, 'rb').read(20)[10:]
				
				if isinstance(item, int):
					self.failUnlessRaises(ValueError, c.path)
				else:
					self.failUnlessRaises(ValueError, c.fd)
				#END handle value error
			#END for each input
			os.close(fd)
		# END for each manager type
			
	def test_memman_operation(self):
		# test more access, force it to actually unmap regions
		fc = FileCreator(self.k_window_test_size, "manager_operation_test")
		data = open(fc.path, 'rb').read()
		fd = os.open(fc.path, os.O_RDONLY)
		max_num_handles = 15
		#small_size = 
		for mtype, args in ( (StaticWindowMapManager, (0, fc.size / 3, max_num_handles)),
							(SlidingWindowMapManager, (fc.size / 100, fc.size / 3, max_num_handles)),):
			for item in (fc.path, fd):
				assert len(data) == fc.size
				
				# small windows, a reasonable max memory. Not too many regions at once
				man = mtype(window_size=args[0], max_memory_size=args[1], max_open_handles=args[2])
				c = man.make_cursor(item)
				
				# still empty (more about that is tested in test_memory_manager()
				assert man.num_open_files() == 0
				assert man.mapped_memory_size() == 0
				
				base_offset = 5000
				# window size is 0 for static managers, hence size will be 0. We take that into consideration
				size = man.window_size() / 2
				assert c.use_region(base_offset, size).is_valid()
				rr = c.region_ref()
				assert rr().client_count() == 2	# the manager and the cursor and us
				
				assert man.num_open_files() == 1
				assert man.num_file_handles() == 1
				assert man.mapped_memory_size() == rr().size()
				
				#assert c.size() == size		# the cursor may overallocate in its static version
				assert c.ofs_begin() == base_offset
				assert rr().ofs_begin() == 0		# it was aligned and expanded
				if man.window_size():
					assert rr().size() == align_to_mmap(man.window_size(), True)	# but isn't larger than the max window (aligned)
				else:
					assert rr().size() == fc.size
				#END ignore static managers which dont use windows and are aligned to file boundaries
				
				assert c.buffer()[:] == data[base_offset:base_offset+(size or c.size())] 
				
				# obtain second window, which spans the first part of the file - it is a still the same window
				nsize = (size or fc.size) - 10
				assert c.use_region(0, nsize).is_valid()
				assert c.region_ref()() == rr()
				assert man.num_file_handles() == 1
				assert c.size() == nsize
				assert c.ofs_begin() == 0
				assert c.buffer()[:] == data[:nsize]
				
				# map some part at the end, our requested size cannot be kept
				overshoot = 4000
				base_offset = fc.size - (size or c.size()) + overshoot
				assert c.use_region(base_offset, size).is_valid()
				if man.window_size():
					assert man.num_file_handles() == 2
					assert c.size() < size
					assert c.region_ref()() is not rr() # old region is still available, but has not curser ref anymore
					assert rr().client_count() == 1 # only held by manager
				else:
					assert c.size() < fc.size
				#END ignore static managers which only have one handle per file
				rr = c.region_ref()
				assert rr().client_count() == 2 # manager + cursor
				assert rr().ofs_begin() < c.ofs_begin() # it should have extended itself to the left
				assert rr().ofs_end() <= fc.size # it cannot be larger than the file
				assert c.buffer()[:] == data[base_offset:base_offset+(size or c.size())]
				
				# unising a region makes the cursor invalid
				c.unuse_region()
				assert not c.is_valid()
				if man.window_size():
					# but doesn't change anything regarding the handle count - we cache it and only 
					# remove mapped regions if we have to
					assert man.num_file_handles() == 2
				#END ignore this for static managers
					
				# iterate through the windows, verify data contents
				# this will trigger map collection after a while
				max_random_accesses = 5000
				num_random_accesses = max_random_accesses
				memory_read = 0
				st = time()
				
				# cache everything to get some more performance
				includes_ofs = c.includes_ofs
				max_mapped_memory_size = man.max_mapped_memory_size()
				max_file_handles = man.max_file_handles()
				mapped_memory_size = man.mapped_memory_size
				num_file_handles = man.num_file_handles
				while num_random_accesses:
					num_random_accesses -= 1
					base_offset = randint(0, fc.size - 1)
					
					# precondition
					if man.window_size():
						assert max_mapped_memory_size >= mapped_memory_size()
					#END statics will overshoot, which is fine
					assert max_file_handles >= num_file_handles()
					assert c.use_region(base_offset, (size or c.size())).is_valid()
					csize = c.size()
					assert c.buffer()[:] == data[base_offset:base_offset+csize]
					memory_read += csize
					
					assert includes_ofs(base_offset)
					assert includes_ofs(base_offset+csize-1)
					assert not includes_ofs(base_offset+csize)
				# END while we should do an access
				elapsed = max(time() - st, 0.001) # prevent zero divison errors on windows
				mb = float(1000 * 1000)
				sys.stderr.write("%s: Read %i mb of memory with %i random on cursor initialized with %s accesses in %fs (%f mb/s)\n" 
								% (mtype, memory_read/mb, max_random_accesses, type(item), elapsed, (memory_read/mb)/elapsed))
		
				# an offset as large as the size doesn't work !
				assert not c.use_region(fc.size, size).is_valid()
				
				# collection - it should be able to collect all
				assert man.num_file_handles()
				assert man.collect()
				assert man.num_file_handles() == 0
			#END for each item
		# END for each manager type
		os.close(fd)



#!/usr/bin/env python
import os
import codecs
try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

import smmap

if os.path.exists("README.rst"):
    long_description = codecs.open('README.rst', "r", "utf-8").read()
else:
    long_description = "See http://github.com/nvie/smmap/tree/master"

setup(
    name="smmap",
    version=smmap.__version__,
    description="A pure git implementation of a sliding window memory map manager",
    author=smmap.__author__,
    author_email=smmap.__contact__,
    url=smmap.__homepage__,
    platforms=["any"],
    license="BSD",
    packages=find_packages(),
    zip_safe=True,
    classifiers=[
        # Picked from
        #    http://pypi.python.org/pypi?:action=list_classifiers
        #"Development Status :: 1 - Planning",
        #"Development Status :: 2 - Pre-Alpha",
        #"Development Status :: 3 - Alpha",
        "Development Status :: 4 - Beta",
        #"Development Status :: 5 - Production/Stable",
        #"Development Status :: 6 - Mature",
        #"Development Status :: 7 - Inactive",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
    ],
    long_description=long_description,
)



"""Module containnig a memory memory manager which provides a sliding window on a number of memory mapped files"""
import os
import sys
import mmap

from mmap import mmap, ACCESS_READ
try:
	from mmap import ALLOCATIONGRANULARITY
except ImportError:
	# in python pre 2.6, the ALLOCATIONGRANULARITY does not exist as it is mainly
	# useful for aligning the offset. The offset argument doesn't exist there though
	from mmap import PAGESIZE as ALLOCATIONGRANULARITY
#END handle pythons missing quality assurance

from sys import getrefcount

__all__ = [	"align_to_mmap", "is_64_bit",
			"MapWindow", "MapRegion", "MapRegionList", "ALLOCATIONGRANULARITY"]

#{ Utilities

def align_to_mmap(num, round_up):
	"""
	Align the given integer number to the closest page offset, which usually is 4096 bytes.
	
	:param round_up: if True, the next higher multiple of page size is used, otherwise
		the lower page_size will be used (i.e. if True, 1 becomes 4096, otherwise it becomes 0)
	:return: num rounded to closest page"""
	res = (num / ALLOCATIONGRANULARITY) * ALLOCATIONGRANULARITY;
	if round_up and (res != num):
		res += ALLOCATIONGRANULARITY
	#END handle size
	return res;
	
def is_64_bit():
	""":return: True if the system is 64 bit. Otherwise it can be assumed to be 32 bit"""
	return sys.maxint > (1<<32) - 1

#}END utilities


#{ Utility Classes 

class MapWindow(object):
	"""Utility type which is used to snap windows towards each other, and to adjust their size"""
	__slots__ = (
				'ofs',		# offset into the file in bytes
				'size'				# size of the window in bytes
				)

	def __init__(self, offset, size):
		self.ofs = offset
		self.size = size

	def __repr__(self):
		return "MapWindow(%i, %i)" % (self.ofs, self.size) 

	@classmethod
	def from_region(cls, region):
		""":return: new window from a region"""
		return cls(region._b, region.size())

	def ofs_end(self):
		return self.ofs + self.size

	def align(self):
		"""Assures the previous window area is contained in the new one"""
		nofs = align_to_mmap(self.ofs, 0)
		self.size += self.ofs - nofs	# keep size constant
		self.ofs = nofs
		self.size = align_to_mmap(self.size, 1)

	def extend_left_to(self, window, max_size):
		"""Adjust the offset to start where the given window on our left ends if possible, 
		but don't make yourself larger than max_size.
		The resize will assure that the new window still contains the old window area"""
		rofs = self.ofs - window.ofs_end()
		nsize = rofs + self.size
		rofs -= nsize - min(nsize, max_size)
		self.ofs = self.ofs - rofs
		self.size += rofs

	def extend_right_to(self, window, max_size):
		"""Adjust the size to make our window end where the right window begins, but don't
		get larger than max_size"""
		self.size = min(self.size + (window.ofs - self.ofs_end()), max_size)


class MapRegion(object):
	"""Defines a mapped region of memory, aligned to pagesizes
	
	**Note:** deallocates used region automatically on destruction"""
	__slots__ = [
					'_b'	, 	# beginning of mapping
					'_mf',	# mapped memory chunk (as returned by mmap)
					'_uc',	# total amount of usages
					'_size', # cached size of our memory map
					'__weakref__'
				]
	_need_compat_layer = sys.version_info[1] < 6
	
	if _need_compat_layer:
		__slots__.append('_mfb')		# mapped memory buffer to provide offset
	#END handle additional slot
	
	#{ Configuration
	# Used for testing only. If True, all data will be loaded into memory at once.
	# This makes sure no file handles will remain open.
	_test_read_into_memory = False
	#} END configuration
	
	
	def __init__(self, path_or_fd, ofs, size, flags = 0):
		"""Initialize a region, allocate the memory map
		:param path_or_fd: path to the file to map, or the opened file descriptor
		:param ofs: **aligned** offset into the file to be mapped 
		:param size: if size is larger then the file on disk, the whole file will be
			allocated the the size automatically adjusted
		:param flags: additional flags to be given when opening the file. 
		:raise Exception: if no memory can be allocated"""
		self._b = ofs
		self._size = 0
		self._uc = 0
		
		if isinstance(path_or_fd, int):
			fd = path_or_fd
		else:
			fd = os.open(path_or_fd, os.O_RDONLY|getattr(os, 'O_BINARY', 0)|flags)
		#END handle fd
		
		try:
			kwargs = dict(access=ACCESS_READ, offset=ofs)
			corrected_size = size
			sizeofs = ofs
			if self._need_compat_layer:
				del(kwargs['offset'])
				corrected_size += ofs
				sizeofs = 0
			# END handle python not supporting offset ! Arg
			
			# have to correct size, otherwise (instead of the c version) it will 
			# bark that the size is too large ... many extra file accesses because
			# if this ... argh !
			actual_size = min(os.fstat(fd).st_size - sizeofs, corrected_size)
			if self._test_read_into_memory:
				self._mf = self._read_into_memory(fd, ofs, actual_size)
			else:
				self._mf = mmap(fd, actual_size, **kwargs)
			#END handle memory mode
			
			self._size = len(self._mf)
			
			if self._need_compat_layer:
				self._mfb = buffer(self._mf, ofs, self._size)
			#END handle buffer wrapping
		finally:
			if isinstance(path_or_fd, basestring):
				os.close(fd)
			#END only close it if we opened it
		#END close file handle
		
	def _read_into_memory(self, fd, offset, size):
		""":return: string data as read from the given file descriptor, offset and size """
		os.lseek(fd, offset, os.SEEK_SET)
		mf = ''
		bytes_todo = size
		while bytes_todo:
			chunk = 1024*1024
			d = os.read(fd, chunk)
			bytes_todo -= len(d)
			mf += d
		#END loop copy items
		return mf
		
	def __repr__(self):
		return "MapRegion<%i, %i>" % (self._b, self.size())
		
	#{ Interface

	def buffer(self):
		""":return: a buffer containing the memory"""
		return self._mf
	
	def map(self):
		""":return: a memory map containing the memory"""
		return self._mf
		
	def ofs_begin(self):
		""":return: absolute byte offset to the first byte of the mapping"""
		return self._b
		
	def size(self):
		""":return: total size of the mapped region in bytes"""
		return self._size
		
	def ofs_end(self):
		""":return: Absolute offset to one byte beyond the mapping into the file"""
		return self._b + self._size
		
	def includes_ofs(self, ofs):
		""":return: True if the given offset can be read in our mapped region"""
		return self._b <= ofs < self._b + self._size
		
	def client_count(self):
		""":return: number of clients currently using this region"""
		# -1: self on stack, -1 self in this method, -1 self in getrefcount
		return getrefcount(self)-3
		
	def usage_count(self):
		""":return: amount of usages so far"""
		return self._uc
		
	def increment_usage_count(self):
		"""Adjust the usage count by the given positive or negative offset"""
		self._uc += 1
		
	# re-define all methods which need offset adjustments in compatibility mode
	if _need_compat_layer:
		def size(self):
			return self._size - self._b
			
		def ofs_end(self):
			# always the size - we are as large as it gets
			return self._size
			
		def buffer(self):
			return self._mfb
			
		def includes_ofs(self, ofs):
			return self._b <= ofs < self._size
	#END handle compat layer
	
	#} END interface
	
	
class MapRegionList(list):
	"""List of MapRegion instances associating a path with a list of regions."""
	__slots__ = (
				'_path_or_fd', 	# path or file descriptor which is mapped by all our regions
				'_file_size'		# total size of the file we map
				)
	
	def __new__(cls, path):
		return super(MapRegionList, cls).__new__(cls)
	
	def __init__(self, path_or_fd):
		self._path_or_fd = path_or_fd
		self._file_size = None
		
	def client_count(self):
		""":return: amount of clients which hold a reference to this instance"""
		return getrefcount(self)-3
		
	def path_or_fd(self):
		""":return: path or file descriptor we are attached to"""
		return self._path_or_fd
		
	def file_size(self):
		""":return: size of file we manager"""
		if self._file_size is None:
			if isinstance(self._path_or_fd, basestring):
				self._file_size = os.stat(self._path_or_fd).st_size
			else:
				self._file_size = os.fstat(self._path_or_fd).st_size
			#END handle path type
		#END update file size
		return self._file_size
	
#} END utilty classes
